//==========================================================================
// Mouse Injector Plugin
//==========================================================================
// Copyright (C) 2017 Carnivorous
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the <organization> nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL CARNIVOROUS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//==========================================================================
#include <stdio.h>
#include <windows.h>
#include "global.h"
#include "device.h"
#include "./manymouse/manymouse.h"
#include "./games/game.h"

//==========================================================================
// Purpose: First called upon device launch
// Changes Globals: alreadyexec, connected
//==========================================================================
int DEV_Init(void)
{
	if(alreadyexec)
		return connected > 1 ? connected : 0;
	alreadyexec = 1;
	connected = ManyMouse_Init();
	int activemice = 0, activekeyboards = 0;
	for(int index = 0; index < connected; index++)
		if(!DEV_Type(index))
			activemice = 1; 
		else
			activekeyboards = 1;
	if(!activemice || !activekeyboards) // fail if mouse or keyboard not detected (mouse and keyboard are required for this plugin)
		connected = 0;
	return connected > 1 ? connected : 0;
}
//==========================================================================
// Purpose: Polls ManyMouse event for input
// Changes Globals: a lot
//==========================================================================
void *DEV_PollInput()
{
	ManyMouseEvent event; // hold current input event (movement, buttons, ect)
	int checkwindowtick = 0; // check if emulator window is in focus
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		for(int button = 0; button < 16; button++) // reset buttons on load
		{
			DEVICE[player].BUTTONPRIM[button] = 0;
			DEVICE[player].BUTTONSEC[button] = 0;
		}
		DEVICE[player].WHEEL = 0;
		togglebuffer = 0;
	}
	while(!stopthread)
	{
		if(mousetoggle)
			SetCursorPos(lockpos.x, lockpos.y); // set mouse position to lock position
		if(togglebuffer > 0)
			togglebuffer--;
		for(int player = PLAYER1; player < ALLPLAYERS; player++)
		{
			DEVICE[player].XPOS = 0; // reset mouse input
			DEVICE[player].YPOS = 0;
			if(DEVICE[player].WHEEL <= 1)  // mouse wheel is not an instant key - treat with cooldown and turn button off once cooled off
			{
				DEVICE[player].WHEEL = 0;
				for(int button = 0; button < 16; button++) // reset wheel scroll once cooled down
				{
					if(PROFILE[player].BUTTONPRIM[button] == VK_WHEELUP || PROFILE[player].BUTTONPRIM[button] == VK_WHEELDOWN || PROFILE[player].BUTTONPRIM[button] == VK_WHEELLEFT || PROFILE[player].BUTTONPRIM[button] == VK_WHEELRIGHT)
						DEVICE[player].BUTTONPRIM[button] = 0;
					if(PROFILE[player].BUTTONSEC[button] == VK_WHEELUP || PROFILE[player].BUTTONSEC[button] == VK_WHEELDOWN || PROFILE[player].BUTTONSEC[button] == VK_WHEELLEFT || PROFILE[player].BUTTONSEC[button] == VK_WHEELRIGHT)
						DEVICE[player].BUTTONSEC[button] = 0;
				}
			}
			else
				DEVICE[player].WHEEL--;
		}
		while(ManyMouse_PollEvent(&event))
		{
			int ignoredevices = 0;
			for(int player = PLAYER1; player < ALLPLAYERS; player++)
			{
				if(PROFILE[player].SETTINGS[CONFIG] == DISABLED) // don't check for disabled players
					continue;
				if(ONLY1PLAYERACTIVE) // if there is only one active player then accept all devices as input
					ignoredevices = 1;
				if(PROFILE[player].SETTINGS[MOUSE] == (int)event.device || ignoredevices) // mouse movement
				{
					if(event.type == MANYMOUSE_EVENT_RELMOTION && mousetoggle)
					{
						if(event.item == 0)
							DEVICE[player].XPOS += event.value;
						else
							DEVICE[player].YPOS += event.value;
					}
					if(event.type == MANYMOUSE_EVENT_BUTTON) // key presses
					{
						for(int button = 0; button < 16; button++)
						{
							int offset = event.item > 1 ? 2 : 1;
							if(PROFILE[player].BUTTONPRIM[button] == (int)event.item + offset) // mouse needs offset to be correct with VK table (left click + offset = 0x01)
								DEVICE[player].BUTTONPRIM[button] = event.value;
							if(PROFILE[player].BUTTONSEC[button] == (int)event.item + offset) // mouse needs offset to be correct with VK table (right click + offset = 0x02)
								DEVICE[player].BUTTONSEC[button] = event.value;
						}
					}
					if(event.type == MANYMOUSE_EVENT_SCROLL) // mouse wheel
					{
						for(int button = 0; button < 16; button++)
						{
							if(event.item == 0) // if VK_WHEELUP/VKWHEELDOWN
							{
								if(event.value > 0 && PROFILE[player].BUTTONPRIM[button] == VK_WHEELUP || event.value < 0 && PROFILE[player].BUTTONPRIM[button] == VK_WHEELDOWN)
									DEVICE[player].BUTTONPRIM[button] = event.value != 0 ? 1 : 0; // filter value to 1 if 1 or -1
								if(event.value > 0 && PROFILE[player].BUTTONSEC[button] == VK_WHEELUP || event.value < 0 && PROFILE[player].BUTTONSEC[button] == VK_WHEELDOWN)
									DEVICE[player].BUTTONSEC[button] = event.value != 0 ? 1 : 0;
							}
							else // if VK_WHEELRIGHT/VKWHEELLEFT
							{
								if(event.value > 0 && PROFILE[player].BUTTONPRIM[button] == VK_WHEELLEFT || event.value < 0 && PROFILE[player].BUTTONPRIM[button] == VK_WHEELRIGHT)
									DEVICE[player].BUTTONPRIM[button] = event.value != 0 ? 1 : 0;
								if(event.value > 0 && PROFILE[player].BUTTONSEC[button] == VK_WHEELLEFT || event.value < 0 && PROFILE[player].BUTTONSEC[button] == VK_WHEELRIGHT)
									DEVICE[player].BUTTONSEC[button] = event.value != 0 ? 1 : 0;
							}
						}
						DEVICE[player].WHEEL = 16; // hold button down for 32ms/64ms (non-oc emu need longer buffer because they run at lower framerate)
					}
				}
				if(event.type == MANYMOUSE_EVENT_KEYBOARD && (PROFILE[player].SETTINGS[KEYBOARD] == (int)event.device || ignoredevices))
				{
					for(int button = 0; button < 16; button++) // check for key presses
					{
						if(PROFILE[player].BUTTONPRIM[button] == (int)event.item)
							DEVICE[player].BUTTONPRIM[button] = event.value;
						if(PROFILE[player].BUTTONSEC[button] == (int)event.item)
							DEVICE[player].BUTTONSEC[button] = event.value;
					}
					if(mousetogglekey == (int)event.item && event.value && !togglebuffer && windowactive) // only toggle on press and if window is active
					{
						mousetoggle = !mousetoggle;
						GetCursorPos(&lockpos);
						togglebuffer = emuoverclock ? 50 : 25; // buffer for 100ms
						if(mouselockonfocus) // if user presses mouse toggle key while lock on focus is on, increase buffer size before mouse is relocked so user can temporarily control mouse to close game/change settings
							togglebuffer *= 15;
					}
				}
			}
		}
		if(checkwindowtick > (emuoverclock ? 250 : 125)) // poll every 500ms
		{
			checkwindowtick = 0;
			if(emulatorwindow != GetForegroundWindow()) // window is inactive
			{
				for(int player = PLAYER1; player < ALLPLAYERS; player++) // reset all input
				{
					DEVICE[player].XPOS = 0;
					DEVICE[player].YPOS = 0;
					for(int button = 0; button < 16; button++)
					{
						DEVICE[player].BUTTONPRIM[button] = 0;
						DEVICE[player].BUTTONSEC[button] = 0;
					}
				}
				GAME_Inject(); // ship empty input to game
				windowactive = 0;
				mousetoggle = 0;
				togglebuffer = 0;
			}
			else // emu window is in focus
			{
				windowactive = 1;
				if(mouselockonfocus && !mousetoggle && !togglebuffer) // if mouselockonfocus is on and window was refocused/mouse toggle is off, turn on mouse toggle
				{
					mousetoggle = 1;
					GetCursorPos(&lockpos);
				}
			}
		}
		else
			checkwindowtick++;
		if(windowactive && GAME_Status()) // if emulator is focused and game is valid
			GAME_Inject(); // inject game
		DEV_Sleep(emuoverclock ? 2 : 4); // 2 for overclocked emus, 4 for normal emus
	}
	return NULL;
}
//==========================================================================
// Purpose: Returns a single key
//==========================================================================
int DEV_ReturnKey(void)
{
	int key = 0;
	ManyMouseEvent event; // hold current input event (movement, buttons, ect)
	while(ManyMouse_PollEvent(&event))
	{
		if(event.type == MANYMOUSE_EVENT_BUTTON)
		{
			if(event.item < 2)
				key = event.item ? VK_RBUTTON : VK_LBUTTON;
			else if(event.item == 2)
				key = VK_MBUTTON;
			else if(event.item == 3)
				key = VK_XBUTTON1;
			else
				key = VK_XBUTTON2;
		}
		else if(event.type == MANYMOUSE_EVENT_SCROLL)
		{
			if(event.item == 0)
				key = event.value > 0 ? VK_WHEELUP : VK_WHEELDOWN;
			else
				key = event.value > 0 ? VK_WHEELRIGHT : VK_WHEELLEFT;
		}
		else if(event.type == MANYMOUSE_EVENT_KEYBOARD)
			key = event.item;
	}
	return key < 0xFF ? key : 0; // if key not recognized send none key
}
//==========================================================================
// Purpose: Returns device id of detected input
//==========================================================================
int DEV_ReturnDeviceID(const int kbflag)
{
	ManyMouseEvent event; // hold current input event (movement, buttons, ect)
	while(ManyMouse_PollEvent(&event))
		if(!kbflag && (event.type == MANYMOUSE_EVENT_BUTTON || event.type == MANYMOUSE_EVENT_SCROLL) || kbflag && event.type == MANYMOUSE_EVENT_KEYBOARD)
			return event.device;
	return -1;
}
//==========================================================================
// Purpose: Wraps ManyMouse device name
//==========================================================================
const char *DEV_Name(const int id)
{
	for(int kbflag = 0; kbflag < 2; kbflag++)
		if(ManyMouse_DeviceName(id, kbflag) != NULL)
			return ManyMouse_DeviceName(id, kbflag);
	return NULL;
}
//==========================================================================
// Purpose: Wraps ManyMouse device name and returns the device type
//==========================================================================
int DEV_Type(const int id)
{
	return ManyMouse_DeviceName(id, 0) != NULL ? 0 : 1; // 0 mouse : 1 keyboard
}
//==========================================================================
// Purpose: Returns the index id within a type (for converting to device combobox)
//==========================================================================
int DEV_TypeIndex(const int id)
{
	int kbcount = -1, mscount = -1;
	for(int index = 0; index < DEV_Init(); index++)
	{
		if(!DEV_Type(index))
			mscount++;
		else
			kbcount++;
		if(index == id)
			return !DEV_Type(id) ? mscount : kbcount;
	}
	return 0;
}
//==========================================================================
// Purpose: Converts a type index id to its real id (for converting from device combobox)
//==========================================================================
int DEV_TypeID(const int id, const int kbflag)
{
	int kbcount = -1, mscount = -1;
	for(int index = 0; index < DEV_Init(); index++)
	{
		if(!DEV_Type(index))
			mscount++;
		else
			kbcount++;
		if(id == mscount && !kbflag || id == kbcount && kbflag)
			return index;
	}
	return 0;
}
//==========================================================================
// Purpose: Pauses whatever for x amount of milliseconds
//==========================================================================
void DEV_Sleep(unsigned int millisec)
{
	Sleep(millisec);
}