//==========================================================================
// Mouse Injector Plugin
//==========================================================================
// Copyright (C) 2017 Carnivorous
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the <organization> nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL CARNIVOROUS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//==========================================================================
#include <math.h>
#include "game.h"
#include "../global.h"

#define TANKXROTATIONLIMIT 6.283185005 // 0x40C90FDA
#define GUNAIMLIMIT 6.879164696 // 0x40DC221E
#define CROSSHAIRLIMIT 5.159373283 // 0x40A51996
#define TIMESTEP (emuoverclock ? 2 : 4) / 1000
// GOLDENEYE ADDRESSES - OFFSET ADDRESSES BELOW (REQUIRES PLAYERBASE TO USE)
#define GE_deathflag 0x800D3038 - 0x800D2FFC
#define GE_camx 0x800D30A8 - 0x800D2FFC
#define GE_camy 0x800D30B8 - 0x800D2FFC
#define GE_fov 0x800D4124 - 0x800D2FFC
#define GE_crosshairx 0x800D3F50 - 0x800D2FFC
#define GE_crosshairy 0x800D3F54 - 0x800D2FFC
#define GE_watch 0x800D3128 - 0x800D2FFC
#define GE_crouchheight 0x800D3000 - 0x800D2FFC
#define GE_gunx 0x800D3F64 - 0x800D2FFC
#define GE_guny 0x800D3F68 - 0x800D2FFC
#define GE_aimingflag 0x800D3084 - 0x800D2FFC
#define GE_currentweapon 0x800D37D0 - 0x800D2FFC
#define GE_multipausemenu 0x800A9D24 - 0x800A73FC
// STATIC ADDRESSES BELOW
#define BONDDATA (unsigned int)EMU_ReadInt(0x8007A0B0, 0) + 0x9C // playable character point (used for all characters)
#define GE_camera 0x80036494 // camera flag (0 = multiplayer, 1 = map overview, 2 = start flyby, 3 = in flyby, 4 = player in control, 5 = trigger restart map)
#define GE_exit 0x800364B0 // exit flag (0 = disable controls, 1 = enable controls)
#define GE_pause 0x80048370 // pause flag (1 = GE is paused)
#define GE_menupage 0x8002A8C0 // menu page id
#define GE_menux 0x8002A908 // crosshair menu cursor x axis
#define GE_menuy 0x8002A90C // crosshair menu cursor y axis
#define GE_tankxrot 0x80036484 // tank x rotation
#define GE_tankflag 0x80036448 // tank flag (0 = walking, 1 = in-tank)
#define GE_matchended 0x8008C700 // multiplayer match flag
#define GE_defaultfov 0x000B78BC // location of field of view default (in rom)

static unsigned int playerbase[4] = {0, 0, 0, 0}; // current player's bonddata address
static int safetoduck[4] = {1, 1, 1, 1}, safetostand[4] = {0, 0, 0, 0}, crouchstance[4] = {0, 0, 0, 0}; // used for crouch toggle changes (limits tick-tocking)
static int crouchingandaiming[4] = {0, 0, 0, 0}; // used for crouch toggle changes (limits tick-tocking)
static float crosshairposx[4], crosshairposy[4], aimx[4], aimy[4];

// external functions/variables from device.c
extern void DEV_Sleep(unsigned int millisec);
// external functions/variables from main.c
extern int emuoverclock;
extern int overridefov;
// external functions/variables from memory.c
extern int EMU_ReadInt(const unsigned int addr, const unsigned int offset);
extern float EMU_ReadFloat(const unsigned int addr, const unsigned int offset);
extern void EMU_WriteInt(const unsigned int addr, const unsigned int offset, const int value);
extern void EMU_WriteFloat(const unsigned int addr, const unsigned int offset, const float value);
extern void EMU_WriteROM(const unsigned int addr, const unsigned int value);

static int GE_Status(void);
static void GE_DetectMap(void);
static void GE_Inject(void);
static void GE_Crouch(const int player);
#define GE_ResetCrouchToggle(X) safetoduck[X] = 1, safetostand[X] = 0, crouchstance[X] = 0 // reset crouch toggle bind
static void GE_AimMode(const int player, const int aimingflag, const float fov);
static void GE_Controller(void);
static void GE_FOVHack(void);
static void GE_SortArray(int *a, const int n);
static void GE_Quit(void);

static const GAMEDRIVER GAMEDRIVER_INTERFACE =
{
	"GoldenEye 007",
	GE_Status,
	GE_Inject,
	GE_Quit
};

const GAMEDRIVER *GAME_GOLDENEYE007 = &GAMEDRIVER_INTERFACE;

//==========================================================================
// Purpose: returns a value, which is then used to check what game is running in game.c
// Q: What is happening here?
// A: We look up some static addresses and if the values are within the expected ranges the program can assume that the game is currently running
//==========================================================================
static int GE_Status(void)
{
	const int ge_camera = EMU_ReadInt(GE_camera, 0), ge_page = EMU_ReadInt(GE_menupage, 0), ge_pause = EMU_ReadInt(GE_pause, 0), ge_exit = EMU_ReadInt(GE_exit, 0);
	const float ge_crosshairx = EMU_ReadFloat(GE_menux, 0), ge_crosshairy = EMU_ReadFloat(GE_menuy, 0);
	if(ge_camera >= 0 && ge_camera <= 10 && ge_page >= -1 && ge_page <= 25 && ge_pause >= 0 && ge_pause <= 1 && ge_exit >= 0 && ge_exit <= 1 && ge_crosshairx >= 20 && ge_crosshairx <= 420 && ge_crosshairy >= 20 && ge_crosshairy <= 310) // if GoldenEye 007 is current game
		return 1;
	return 0;
}
//==========================================================================
// Purpose: searches for the current map memory locations
// Changes Globals: playerbase, safetoduck, safetostand, crouchstance
//==========================================================================
static void GE_DetectMap(void)
{
	if(playerbase[PLAYER1] != BONDDATA && (EMU_ReadInt(GE_camera, 0) == 1 || EMU_ReadInt(GE_camera, 0) == 4 || EMU_ReadInt(GE_camera, 0) == 9))
	{
		unsigned int pointerdump[64];
		for(int player = PLAYER1; player < ALLPLAYERS; player++)
			GE_ResetCrouchToggle(player); // reset crouch toggle on new map
		for(int index = 0; index < 64; index++)
		{
			pointerdump[index] = BONDDATA;
			DEV_Sleep(1);
		}
		GE_SortArray((int *)pointerdump, sizeof pointerdump / sizeof pointerdump[0]);
		playerbase[PLAYER1] = pointerdump[0];
		int playerfound = PLAYER1;
		for(int index = 0; index < 64; index++)
		{
			if(playerbase[PLAYER1] != pointerdump[index] && playerfound == PLAYER1)
			{
				playerbase[PLAYER2] = pointerdump[index];
				playerfound = PLAYER2;
			}
			else if(playerbase[PLAYER2] != pointerdump[index] && playerfound == PLAYER2)
			{
				playerbase[PLAYER3] = pointerdump[index];
				playerfound = PLAYER3;
			}
			else if(playerbase[PLAYER3] != pointerdump[index] && playerfound == PLAYER3)
			{
				playerbase[PLAYER4] = pointerdump[index];
				break;
			}
		}
	}
}
//==========================================================================
// Purpose: calculate mouse movement and inject into current game
// Changes Globals: safetoduck, safetostand, crouchstance
// Q: Why is aiming mode disabled for non-overclocked emulators?
// A: If the emulator is not overclocked the crosshair cannot be locked perfectly. The result will look 'twitchy' and you will see some visible injection artifacts with the crosshair. It's better to just disable aiming mode altogether for non-overclocked emulators
// Q: Could you explain !aimingflag ? 10.0f : 40.0f
// A: While the player is aiming weapon sway will be reduced by 75%. While this is not in the original design of GE/PD I feel it gives a legitimate advantage for aiming instead of only making the crosshair visible
// Q: Could you explain fov = !aimingflag ? EMU_ReadFloat(playerbase[player], GE_fov) : fmin(EMU_ReadFloat(playerbase[player], GE_fov), 60)
// A: While having a custom fov it might be larger than 60. When aiming with a weapon the fov will zoom back down to 60 and during that time it cause the mouse sensitivity to be multiplied (we divide by 60, see answer below). This is annoying and throws your aiming off so we limit the fov to always be under 60 when aiming
// Q: Could you explain basefov = !aimingflag ? (float)overridefov : 60
// A: Aiming sets your field of view to 60 or lower, it overrides whatever custom fov you may have set so use 60 as a base if aiming
//==========================================================================
static void GE_Inject(void)
{
	GE_DetectMap();
	if(EMU_ReadInt(GE_menupage, 0) < 1 && overridefov != 60) // fov can only be injected at boot sequence before code blocks are cached, so inject until the main menu
		GE_FOVHack();
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		if(PROFILE[player].SETTINGS[CONFIG] == DISABLED) // don't compute for disabled players
			continue;
		const int camera = EMU_ReadInt(GE_camera, 0), dead = EMU_ReadInt(playerbase[player], GE_deathflag), watch = EMU_ReadInt(playerbase[player], GE_watch), exit = EMU_ReadInt(GE_exit, 0), pause = EMU_ReadInt(GE_pause, 0), aimingflag = EMU_ReadInt(playerbase[player], GE_aimingflag), menupage = EMU_ReadInt(GE_menupage, 0), tankflag = EMU_ReadInt(GE_tankflag, 0), mproundend = EMU_ReadInt(GE_matchended, 0), mppausemenu = EMU_ReadInt(playerbase[player], GE_multipausemenu), cursoraimingflag = PROFILE[player].SETTINGS[GEAIMMODE] && emuoverclock && aimingflag ? 1 : 0;
		const float fov = !aimingflag ? EMU_ReadFloat(playerbase[player], GE_fov) : fmin(EMU_ReadFloat(playerbase[player], GE_fov), 60), basefov = !aimingflag ? (float)overridefov : 60, mouseaccel = PROFILE[player].SETTINGS[ACCELERATION] ? sqrt(DEVICE[player].XPOS * DEVICE[player].XPOS + DEVICE[player].YPOS * DEVICE[player].YPOS) / (float)(emuoverclock ? 2 : 4) / 12.f * (float)(PROFILE[player].SETTINGS[ACCELERATION] + 1) : 0, sensitivity = PROFILE[player].SETTINGS[SENSITIVITY] / 40.0f * fmax(mouseaccel, 1), gunsensitivity = sensitivity * (PROFILE[player].SETTINGS[CROSSHAIR] / 2.5f);
		float camx = EMU_ReadFloat(playerbase[player], GE_camx), camy = EMU_ReadFloat(playerbase[player], GE_camy);
		if(camx >= 0 && camx <= 360 && camy >= -90 && camy <= 90 && fov >= 7 && fov <= 120 && dead == 0 && watch == 0 && pause == 0 && (camera == 4 || camera == 0) && exit == 1 && menupage == 11 && !mproundend && !mppausemenu) // if safe to inject
		{
			GE_AimMode(player, cursoraimingflag, fov);
			if(!tankflag) // player is on foot
			{
				GE_Crouch(player); // only allow crouching if player is not in tank
				if(!cursoraimingflag) // if not aiming (geaimmode is off/emulator is not overclocked)
					camx += DEVICE[player].XPOS / 10.0f * sensitivity * (fov / basefov); // regular mouselook calculation
				else
					camx += aimx[player] * (fov / basefov); // scroll screen with aimx/aimy
				if(camx < 0)
					camx += 360;
				else if(camx >= 360)
					camx -= 360;
				EMU_WriteFloat(playerbase[player], GE_camx, camx);
			}
			else // player is in tank
			{
				GE_ResetCrouchToggle(player); // reset crouch toggle if in tank
				float tankx = EMU_ReadFloat(GE_tankxrot, 0);
				if(!cursoraimingflag)
					tankx += DEVICE[player].XPOS / 10.0f * sensitivity / (360 / TANKXROTATIONLIMIT * 2.5) * (fov / basefov);
				else
					tankx += aimx[player] / (360 / TANKXROTATIONLIMIT * 2.5) * (fov / basefov);
				if(tankx < 0)
					tankx += TANKXROTATIONLIMIT;
				else if(tankx >= TANKXROTATIONLIMIT)
					tankx -= TANKXROTATIONLIMIT;
				EMU_WriteFloat(GE_tankxrot, 0, tankx);
			}
			if(!cursoraimingflag)
				camy += (!PROFILE[player].SETTINGS[INVERTPITCH] ? -DEVICE[player].YPOS : DEVICE[player].YPOS) / 10.0f * sensitivity * (fov / basefov);
			else
				camy += -aimy[player] * (fov / basefov);
			if(camy < -20 && tankflag) // tank limits player from looking down -20
				camy = -20;
			if(camy < -90)
				camy = -90;
			else if(camy > 90)
				camy = 90;
			EMU_WriteFloat(playerbase[player], GE_camy, camy);
			if(PROFILE[player].SETTINGS[CROSSHAIR] && !cursoraimingflag) // if crosshair movement is enabled (do not bother calculating weapon movement while the player is in aim mode)
			{
				if(!tankflag)
				{
					float gunx = EMU_ReadFloat(playerbase[player], GE_gunx), crosshairx = EMU_ReadFloat(playerbase[player], GE_crosshairx); // after camera x and y have been calculated and injected, calculate the gun/crosshair movement
					gunx += DEVICE[player].XPOS / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / basefov) * 0.019f;
					crosshairx += DEVICE[player].XPOS / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / 4 / (basefov / 4)) * 0.01912f;
					if(gunx < -GUNAIMLIMIT)
						gunx = -GUNAIMLIMIT;
					else if(gunx > GUNAIMLIMIT)
						gunx = GUNAIMLIMIT;
					if(crosshairx < -CROSSHAIRLIMIT)
						crosshairx = -CROSSHAIRLIMIT;
					else if(crosshairx > CROSSHAIRLIMIT)
						crosshairx = CROSSHAIRLIMIT;
					EMU_WriteFloat(playerbase[player], GE_gunx, gunx);
					EMU_WriteFloat(playerbase[player], GE_crosshairx, crosshairx);
				}
				if((!tankflag && camy > -90 || tankflag && camy > -20) && camy < 90) // only allow player's gun to pitch within a valid range
				{
					float guny = EMU_ReadFloat(playerbase[player], GE_guny), crosshairy = EMU_ReadFloat(playerbase[player], GE_crosshairy);
					guny += (!PROFILE[player].SETTINGS[INVERTPITCH] ? DEVICE[player].YPOS : -DEVICE[player].YPOS) / (!aimingflag ? 40.0f : 20.0f) * gunsensitivity * (fov / basefov) * 0.025f;
					crosshairy += (!PROFILE[player].SETTINGS[INVERTPITCH] ? DEVICE[player].YPOS : -DEVICE[player].YPOS) / (!aimingflag ? 40.0f : 20.0f) * gunsensitivity * (fov / 4 / (basefov / 4)) * 0.0225f;
					if(guny < -GUNAIMLIMIT)
						guny = -GUNAIMLIMIT;
					else if(guny > GUNAIMLIMIT)
						guny = GUNAIMLIMIT;
					if(crosshairy < -CROSSHAIRLIMIT)
						crosshairy = -CROSSHAIRLIMIT;
					else if(crosshairy > CROSSHAIRLIMIT)
						crosshairy = CROSSHAIRLIMIT;
					EMU_WriteFloat(playerbase[player], GE_guny, guny);
					EMU_WriteFloat(playerbase[player], GE_crosshairy, crosshairy);
				}
			}
		}
		else if(player == PLAYER1 && menupage != 11 && menupage != 23) // if user is in menu (only payer 1 can control menu)
		{
			float menucrosshairx = EMU_ReadFloat(GE_menux, 0), menucrosshairy = EMU_ReadFloat(GE_menuy, 0);
			menucrosshairx += DEVICE[player].XPOS / 10.0f * sensitivity * 6;
			menucrosshairy += DEVICE[player].YPOS / 10.0f * sensitivity * (400.0f / 290.0f * 6); // y is a little weaker then x in the menu so add more power to make it feel even with x axis
			if(menucrosshairx < 20)
				menucrosshairx = 20;
			else if(menucrosshairx > 420)
				menucrosshairx = 420;
			if(menucrosshairy < 20)
				menucrosshairy = 20;
			else if(menucrosshairy > 310)
				menucrosshairy = 310;
			EMU_WriteFloat(GE_menux, 0, menucrosshairx);
			EMU_WriteFloat(GE_menuy, 0, menucrosshairy);
		}
		if(dead || menupage != 11) // if player is dead or in menu, reset crouch toggle
			GE_ResetCrouchToggle(player);
	}
	GE_Controller(); // set controller data
}
//==========================================================================
// Purpose: crouching function for GoldenEye (2 = stand, 0 = duck)
// Changes Globals: safetoduck, crouchstance, safetostand, crouchingandaiming
// Q: Can you explain the long if else statement?
// A: When the player is holding down crouch while aiming, GE will force the player to stand as part of its auto stand design. The code below combats this by setting the correct button combinations required to crouch and stand without any interference from auto stand (crouchingandaiming[player]). Read below for a more complex explanation.
// Pressing R will make the player stand and pressing R + c-down will make the player crouch. If the player is pressing CTRL while in aiming mode then the below code will turn on c-down, which will make the player crouch (this is to combat the auto stand design while pressing R). The only issue with this technique (besides from being ugly) is when the player has the sniper rifle equipped, c-up and c-down are instead used to control the zoom for the sniper rifle. This means c-down will make the sniper rifle zoom out instead of making the player crouch. The sniper rifle is the only weapon which does not require c-down to make the player crouch (playerbase = 0 works fine). A cheap way around this is to disable the c-down method when the current weapon id matches the sniper rifle id (17)
//==========================================================================
static void GE_Crouch(const int player)
{
	int crouchheld = DEVICE[player].BUTTONPRIM[CROUCH] || DEVICE[player].BUTTONSEC[CROUCH];
	if(PROFILE[player].SETTINGS[CROUCHTOGGLE]) // check and change player stance
	{
		if(safetoduck[player] && crouchheld) // standing to ducking
			safetoduck[player] = 0, crouchstance[player] = 1;
		if(!safetoduck[player] && !crouchheld) // crouch is no longer being held, ready to stand
			safetostand[player] = 1;
		if(safetostand[player] && crouchheld) // standing up
			safetoduck[player] = 1, crouchstance[player] = 0;
		if(safetostand[player] && safetoduck[player] && !crouchheld) // crouch key not active, ready to change toggle
			safetostand[player] = 0;
		crouchheld = crouchstance[player];
	}
	if(crouchheld && EMU_ReadInt(playerbase[player], GE_aimingflag) && EMU_ReadInt(playerbase[player], GE_currentweapon) != 17) // player is holding down crouch and aiming so set c-down (see FAQ for more information)
		crouchingandaiming[player] = 1;
	else
		crouchingandaiming[player] = 0;
	if(crouchheld) // user is holding down the crouch button
		EMU_WriteInt(playerbase[player], 0, 0); // set in-game crouch flag to lowest point
	else // player isn't holding down crouch
		EMU_WriteInt(playerbase[player], 0, 2); // force standing flag value (the side effect to this brute force method is it causes a speed glitch while crawling in tunnels in PD only)
}
//==========================================================================
// Purpose: replicate the original aiming system for GE, uses aimx/y to move screen when crosshair is on border of screen
// Changes Globals: crosshairposx, crosshairposy, aimx, aimy
//==========================================================================
static void GE_AimMode(const int player, const int aimingflag, const float fov)
{
	const float crosshairx = EMU_ReadFloat(playerbase[player], GE_crosshairx), crosshairy = EMU_ReadFloat(playerbase[player], GE_crosshairy), offsetpos[2][33] = {{0, 0, 0, 0, 0.125, 0.125, 0.15, 0.5, 0.8, 0.4, 0.5, 0.5, 0.55, 0.9, 0.25, 0.625, 0.65, 0.85, 0.25, 0.15, 0.125, 0.125, 0.5, 1.3, 1, 1.05, 0, 0, 0, 0, 1.3, 1.7}, {0, 0, 0, 0, 0.1, 0.1, 0.2, 0.325, 1, 0.3, 0.425, 0.425, 0.525, 0.95, 0.1, 0.55, 0.9, 0.8, 0.25, 0.1, 0.1, 0.1, 0.275, 1.925, 1, 1.3, 0, 0, 0, 0, 0.8, 0.65}}; // table of X/Y offset for weapons
	const int currentweapon = EMU_ReadInt(playerbase[player], GE_currentweapon);
	if(aimingflag) // if player is aiming
	{
		crosshairposx[player] += DEVICE[player].XPOS / 10.0f * (PROFILE[player].SETTINGS[SENSITIVITY] / 292.0f); // calculate the crosshair position
		crosshairposy[player] += (!PROFILE[player].SETTINGS[INVERTPITCH] ? DEVICE[player].YPOS : -DEVICE[player].YPOS) / 10.0f * (PROFILE[player].SETTINGS[SENSITIVITY] / 292.0f);
		if(crosshairposx[player] < -CROSSHAIRLIMIT) // apply limits then inject
			crosshairposx[player] = -CROSSHAIRLIMIT;
		else if(crosshairposx[player] > CROSSHAIRLIMIT)
			crosshairposx[player] = CROSSHAIRLIMIT;
		if(crosshairposy[player] < -CROSSHAIRLIMIT)
			crosshairposy[player] = -CROSSHAIRLIMIT;
		else if(crosshairposy[player] > CROSSHAIRLIMIT)
			crosshairposy[player] = CROSSHAIRLIMIT;
		EMU_WriteFloat(playerbase[player], GE_crosshairx, crosshairposx[player]);
		EMU_WriteFloat(playerbase[player], GE_crosshairy, crosshairposy[player]);
		EMU_WriteFloat(playerbase[player], GE_gunx, crosshairposx[player] * (1.11f + (currentweapon >= 0 && currentweapon <= 32 ? offsetpos[0][currentweapon] : 0.15f)) + (fov / 60) - 1); // calculate and inject the gun angles (uses pre-made pos table or if unknown weapon use fail-safe value)
		EMU_WriteFloat(playerbase[player], GE_guny, crosshairposy[player] * (1.11f + (currentweapon >= 0 && currentweapon <= 32 ? offsetpos[1][currentweapon] : 0)) + (fov / 60) - 1);
		if(crosshairx > 0 && crosshairx / CROSSHAIRLIMIT > 0.9f) // if crosshair is within the last 90% of the border then calculate a linear scrolling speed and enable mouselook
			aimx[player] = (crosshairx / CROSSHAIRLIMIT - 0.9f) * 1500 * TIMESTEP;
		else if(crosshairx < 0 && crosshairx / CROSSHAIRLIMIT < -0.9f)
			aimx[player] = (crosshairx / CROSSHAIRLIMIT + 0.9f) * 1500 * TIMESTEP;
		else
			aimx[player] = 0;
		if(crosshairy > 0 && crosshairy / CROSSHAIRLIMIT > 0.9f)
			aimy[player] = (crosshairy / CROSSHAIRLIMIT - 0.9f) * 1500 * TIMESTEP;
		else if(crosshairy < 0 && crosshairy / CROSSHAIRLIMIT < -0.9f)
			aimy[player] = (crosshairy / CROSSHAIRLIMIT + 0.9f) * 1500 * TIMESTEP;
		else
			aimy[player] = 0;
	}
	else // player is not aiming so reset crosshairposxy values
		crosshairposx[player] = crosshairx, crosshairposy[player] = crosshairy;
}
//==========================================================================
// Purpose: calculate and send emulator key combo
//==========================================================================
static void GE_Controller(void)
{
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		Controller[player].U_CBUTTON = DEVICE[player].BUTTONPRIM[FORWARDS] || DEVICE[player].BUTTONSEC[FORWARDS];
		Controller[player].D_CBUTTON = DEVICE[player].BUTTONPRIM[BACKWARDS] || DEVICE[player].BUTTONSEC[BACKWARDS] || crouchingandaiming[player];
		Controller[player].L_CBUTTON = DEVICE[player].BUTTONPRIM[STRAFELEFT] || DEVICE[player].BUTTONSEC[STRAFELEFT];
		Controller[player].R_CBUTTON = DEVICE[player].BUTTONPRIM[STRAFERIGHT] || DEVICE[player].BUTTONSEC[STRAFERIGHT];
		Controller[player].Z_TRIG = DEVICE[player].BUTTONPRIM[FIRE] || DEVICE[player].BUTTONSEC[FIRE] || DEVICE[player].BUTTONPRIM[PREVIOUSWEAPON] || DEVICE[player].BUTTONSEC[PREVIOUSWEAPON];
		Controller[player].R_TRIG = DEVICE[player].BUTTONPRIM[AIM] || DEVICE[player].BUTTONSEC[AIM];
		Controller[player].A_BUTTON = DEVICE[player].BUTTONPRIM[ACCEPT] || DEVICE[player].BUTTONSEC[ACCEPT] || DEVICE[player].BUTTONPRIM[PREVIOUSWEAPON] || DEVICE[player].BUTTONSEC[PREVIOUSWEAPON] || DEVICE[player].BUTTONPRIM[NEXTWEAPON] || DEVICE[player].BUTTONSEC[NEXTWEAPON];
		Controller[player].B_BUTTON = DEVICE[player].BUTTONPRIM[CANCEL] || DEVICE[player].BUTTONSEC[CANCEL];
		Controller[player].START_BUTTON = DEVICE[player].BUTTONPRIM[START] || DEVICE[player].BUTTONSEC[START];
		DEVICE[player].ARROW[0] = (DEVICE[player].BUTTONPRIM[UP] || DEVICE[player].BUTTONSEC[UP]) ? 120 : 0;
		DEVICE[player].ARROW[1] = (DEVICE[player].BUTTONPRIM[DOWN] || DEVICE[player].BUTTONSEC[DOWN]) ? -120 : 0;
		DEVICE[player].ARROW[2] = (DEVICE[player].BUTTONPRIM[LEFT] || DEVICE[player].BUTTONSEC[LEFT]) ? -120 : 0;
		DEVICE[player].ARROW[3] = (DEVICE[player].BUTTONPRIM[RIGHT] || DEVICE[player].BUTTONSEC[RIGHT]) ? 120 : 0;
		Controller[player].X_AXIS = DEVICE[player].ARROW[0] + DEVICE[player].ARROW[1];
		Controller[player].Y_AXIS = DEVICE[player].ARROW[2] + DEVICE[player].ARROW[3];
	}
	if(EMU_ReadInt(GE_menupage, 0) != 11 && !Controller[PLAYER1].B_BUTTON) // pressing aim will act like the back button (only in menus and for player 1)
		Controller[PLAYER1].B_BUTTON = DEVICE[PLAYER1].BUTTONPRIM[AIM] || DEVICE[PLAYER1].BUTTONSEC[AIM];
}
//==========================================================================
// Purpose: override default fov
//==========================================================================
static void GE_FOVHack(void)
{
	float newfov = overridefov;
	unsigned int unsignedinteger = *(unsigned int *)(float *)(&newfov);
	EMU_WriteROM(GE_defaultfov, 0x3C010000 + (short)(unsignedinteger / 0x10000));
}
//==========================================================================
// Purpose: sorting function for pointer scanning
//==========================================================================
static void GE_SortArray(int *a, const int n)
{
	if(n < 2)
		return;
	int p = a[n / 2];
	int *l = a;
	int *r = a + n - 1;
	while(l <= r)
	{
		while(*l < p)
			l++;
		while(*r > p)
			r--;
		if(l <= r)
		{
			int t = *l;
			*l++ = *r;
			*r-- = t;
		}
	}
	GE_SortArray(a, r - a + 1);
	GE_SortArray(l, a + n - l);
}
//==========================================================================
// Purpose: run when emulator closes rom
// Changes Globals: playerbase, safetoduck, safetostand, crouchstance, crouchingandaiming
//==========================================================================
static void GE_Quit(void)
{
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		playerbase[player] = 0;
		GE_ResetCrouchToggle(player);
		crouchingandaiming[player] = 0;
	}
}