//==========================================================================
// Mouse Injector Plugin
//==========================================================================
// Copyright (C) 2017 Carnivorous
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the <organization> nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL CARNIVOROUS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//==========================================================================
#include <math.h>
#include "game.h"
#include "../global.h"

#define GUNAIMLIMIT 14.12940025 // 0x41621206
#define CROSSHAIRLIMIT 18.76135635 // 0x41961742
#define BIKEXROTATIONLIMIT 6.282184601 // 0x40C907A8
#define BIKEROLLLIMIT 0.7852724195 // 0xBF49079D
// PERFECT DARK ADDRESSES - OFFSET ADDRESSES BELOW (REQUIRES PLAYERBASE/BIKEBASE TO USE)
#define PD_deathflag 0x801BB778 - 0x801BB74C
#define PD_camx 0x801BB7E4 - 0x801BB74C
#define PD_camy 0x801BB7F4 - 0x801BB74C
#define PD_fov 0x801BCEE8 - 0x801BB74C
#define PD_crosshairx 0x801BCD08 - 0x801BB74C
#define PD_crosshairy 0x801BCD0C - 0x801BB74C
#define PD_grabflag 0x801BB850 - 0x801BB74C
#define PD_gunrx 0x801BC374 - 0x801BB74C
#define PD_gunry 0x801BC378 - 0x801BB74C
#define PD_gunlx 0x801BCB18 - 0x801BB74C
#define PD_gunly 0x801BCB1C - 0x801BB74C
#define PD_aimingflag 0x801BB7C0 - 0x801BB74C
#define PD_bikeroll 0x8052D69C - 0x8052D64C
// STATIC ADDRESSES BELOW
#define JOANNADATA(X) (unsigned int)EMU_ReadInt(0x8009A024, X * 0x4) + 0xAC // player pointer address (0x4 offset for other players)
#define PD_menu(X) 0x80070750 + X * 0x4 // player menu flag (0 = PD is in menu) (0x4 offset for other players)
#define PD_camera 0x8009A26C // camera flag (1 = gameplay, 2 & 3 = ???, 4 = multiplayer sweep, 5 = gameover screen, 6 = cutscene mode)
#define PD_pause 0x80084014 // menu flag (1 = PD is paused)
#define PD_menuitem 0x800739F8 // menu item flag (used to check if PD is running)
#define PD_tickrate 0x80099FC0 // game loop's tickrate (used to detect when PD is running at lower framerate - PD was designed to halve game tickrate when detected bottlenecks)
#define PD_timer 0x8008DBD0 // we intentionally screw with this timer address to force PD to run at unlocked speed
#define PD_speed 0x8009A424 // used to check if the game is slow motion or normal speed
#define PD_mpspeed 0x800ACB91 // used to check if the match is set to slow motion or normal speed
#define PD_mppause 0x800ACBA2 // used to check if multiplayer match is paused
#define PD_slowmocheat 0x800A21D5 // slow motion single player cheat
#define PD_officialrom 0x801AD61C // location of text shown in intro (used to determine if user is playing romhack)
#define PD_defaultfov 0x802EAA5C // location of field of view default

static unsigned int playerbase[4] = {0, 0, 0, 0}; // current player's joannadata address
static unsigned int bikebase[4][2] = {{0, 0}, {0, 0}, {0, 0}, {0, 0}}; // hoverbike's address and player's last grab state (used to find the exact moment when the player hops on a bike)
static float xmenu[4] = {0, 0, 0, 0}, ymenu[4] = {0, 0, 0, 0}; // for pd radial nav function
static int radialmenudirection[4][4] = {{0, 0, 0, 0}, {0, 0, 0, 0}}; // used to override c buttons if user is interacting with a radial menu
static int safetoduck[4] = {1, 1, 1, 1}, safetostand[4] = {0, 0, 0, 0}, crouchstance[4] = {0, 0, 0, 0}; // used for crouch toggle changes (limits tick-tocking)
static int limitspeedhack = 0; // self-explanatory

// external functions/variables from device.c
extern void DEV_Sleep(unsigned int millisec);
// external functions/variables from main.c
extern int emuoverclock;
extern int pdspeedhack;
extern int overridefov;
// external functions/variables from memory.c
extern int EMU_ReadInt(const unsigned int addr, const unsigned int offset);
extern float EMU_ReadFloat(const unsigned int addr, const unsigned int offset);
extern void EMU_WriteInt(const unsigned int addr, const unsigned int offset, const int value);
extern void EMU_WriteFloat(const unsigned int addr, const unsigned int offset, const float value);

static int PD_Status(void);
static void PD_DetectMap(void);
static void PD_Inject(void);
static void PD_Crouch(const int player);
#define PD_ResetCrouchToggle(X) safetoduck[X] = 1, safetostand[X] = 0, crouchstance[X] = 0 // reset crouch toggle bind
#define PD_ResetRadialMenuBtns(X) for(int direction = 0; direction < 4; direction++) radialmenudirection[X][direction] = 0 // reset direction buttons
static void PD_RadialMenuNav(const int player);
static void PD_Controller(void);
static void PD_FOVHack(void);
static void PD_SpeedHack(void);
static void PD_Quit(void);

static const GAMEDRIVER GAMEDRIVER_INTERFACE =
{
	"Perfect Dark",
	PD_Status,
	PD_Inject,
	PD_Quit
};

const GAMEDRIVER *GAME_PERFECTDARK = &GAMEDRIVER_INTERFACE;

//==========================================================================
// Purpose: returns a value, which is then used to check what game is running in game.c
// Q: What is happening here?
// A: We look up some static addresses and if the values are within the expected ranges the program can assume that the game is currently running
//==========================================================================
static int PD_Status(void)
{
	const int pd_menu = EMU_ReadInt(PD_menu(0), 0), pd_camera = EMU_ReadInt(PD_camera, 0), pd_pause = EMU_ReadInt(PD_pause, 0), pd_romcheck = EMU_ReadInt(PD_menuitem, 0);
	if(pd_menu >= 0 && pd_menu <= 1 && pd_camera >= 0 && pd_camera <= 7 && pd_pause >= 0 && pd_pause <= 1 && pd_romcheck == 0x04010000)
		return 1;
	return 0;
}
//==========================================================================
// Purpose: searches for the current map memory locations
// Changes Globals: playerbase, safetoduck, safetostand, crouchstance, xmenu, ymenu, bikebase
//==========================================================================
static void PD_DetectMap(void)
{
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		if(playerbase[player] != JOANNADATA(player) && (EMU_ReadInt(PD_camera, 0) == 1 || EMU_ReadInt(PD_camera, 0) == 3 || EMU_ReadInt(PD_camera, 0) == 4 || EMU_ReadInt(PD_camera, 0) == 6))
		{
			playerbase[player] = JOANNADATA(player);
			PD_ResetCrouchToggle(player); // reset crouch toggle on new map
			xmenu[player] = 0; // reset menu direction
			ymenu[player] = 0;
		}
		if(EMU_ReadInt(playerbase[player], PD_grabflag) == 3 && bikebase[player][1] != 3) // player has hopped on a bike, search for bike pointer
		{
			const unsigned int bikepointers[14] = {0x8052D5DC, 0x8054B1E8, 0x80552D8C, 0x8052A4FC, 0x8054ECDC, 0x80556DF4, 0x80501438, 0x8050EED4, 0x805142A4, 0x804CA4D4, 0x804D10D4, 0x803FF4F0, 0x803FF6A4, 0x803FF7BC};
			unsigned int bikedump[64][14];
			for(int i = 0; i < 64; i++)
			{
				for(int j = 0; j < 14; j++)
				{
					bikedump[i][j] = EMU_ReadInt(bikepointers[j], 0);
					if(j > 1 && bikedump[i][j] > 0x80000000 && bikedump[i][j] < 0x80800000 && EMU_ReadFloat(bikedump[i][j] + 0x6C, 0) <= BIKEXROTATIONLIMIT && EMU_ReadFloat(bikedump[i][j] + 0x6C, 0) >= -BIKEXROTATIONLIMIT && EMU_ReadFloat(bikedump[i][j] + 0x6C, PD_bikeroll) <= BIKEROLLLIMIT && EMU_ReadFloat(bikedump[i][j] + 0x6C, PD_bikeroll) >= -BIKEROLLLIMIT && (bikedump[i][j] == bikedump[i][j - 1] || bikedump[i][j] == bikedump[i][j - 2] || bikedump[i][j - 1] == bikedump[i][j - 2]))
					{
						bikebase[player][0] = bikedump[i][j] + 0x6C;
						break;
					}
				}
				if(bikebase[player][0])
					break;
				else
					DEV_Sleep(1);
			}
		}
		else if(bikebase[player][1] != 3)
			bikebase[player][0] = 0;
		bikebase[player][1] = EMU_ReadInt(playerbase[player], PD_grabflag); // remember player's last grab state (used to find the exact moment when the player hops on a bike)
	}
}
//==========================================================================
// Purpose: calculate mouse movement and inject into current game
// Changes Globals: safetoduck, safetostand, crouchstance
//==========================================================================
static void PD_Inject(void)
{
	PD_DetectMap();
	if(emuoverclock && pdspeedhack)
		PD_SpeedHack();
	if(overridefov != 60 && !playerbase[PLAYER1]) // fov can only be injected at boot sequence before code blocks are cached, so inject until player has spawned
		PD_FOVHack();
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		if(PROFILE[player].SETTINGS[CONFIG] == DISABLED) // don't compute for disabled players
			continue;
		const int camera = EMU_ReadInt(PD_camera, 0), dead = EMU_ReadInt(playerbase[player], PD_deathflag), menu = EMU_ReadInt(PD_menu(player), 0), pause = EMU_ReadInt(PD_pause, 0), mppause = EMU_ReadInt(PD_mppause, 0), aimingflag = EMU_ReadInt(playerbase[player], PD_aimingflag), grabflag = EMU_ReadInt(playerbase[player], PD_grabflag);
		const float fov = !aimingflag ? EMU_ReadFloat(playerbase[player], PD_fov) : fmin(EMU_ReadFloat(playerbase[player], PD_fov), 60), basefov = !aimingflag ? (float)overridefov : 60, mouseaccel = PROFILE[player].SETTINGS[ACCELERATION] ? sqrt(DEVICE[player].XPOS * DEVICE[player].XPOS + DEVICE[player].YPOS * DEVICE[player].YPOS) / (float)(emuoverclock ? 2 : 4) / 12.f * (float)(PROFILE[player].SETTINGS[ACCELERATION] + 1) : 0, sensitivity = PROFILE[player].SETTINGS[SENSITIVITY] / 40.0f * fmax(mouseaccel, 1), gunsensitivity = sensitivity * (PROFILE[player].SETTINGS[CROSSHAIR] / 2.5f);
		float camx = EMU_ReadFloat(playerbase[player], PD_camx), camy = EMU_ReadFloat(playerbase[player], PD_camy), bikex = EMU_ReadFloat(bikebase[player][0], 0), bikeroll = EMU_ReadFloat(bikebase[player][0], PD_bikeroll);
		if(camx >= 0 && camx <= 360 && camy >= -90 && camy <= 90 && fov >= 2 && fov <= 120 && dead == 0 && menu == 1 && pause == 0 && mppause < 16777476 && camera == 1 && (grabflag == 0 || grabflag == 4 || grabflag == 3 && bikebase[player][0] && bikex <= BIKEXROTATIONLIMIT && bikex >= -BIKEXROTATIONLIMIT && bikeroll <= BIKEROLLLIMIT && bikeroll >= -BIKEROLLLIMIT)) // if safe to inject
		{
			if(grabflag != 3) // if player is on foot
			{
				PD_Crouch(player);
				camx += DEVICE[player].XPOS / 10.0f * sensitivity * (fov / basefov);
				if(camx < 0)
					camx += 360;
				else if(camx >= 360)
					camx -= 360;
				EMU_WriteFloat(playerbase[player], PD_camx, camx);
			}
			else // if player is riding hoverbike (and hoverbike address is valid)
			{
				PD_ResetCrouchToggle(player);
				bikex -= DEVICE[player].XPOS / 10.0f * sensitivity / (360 / BIKEXROTATIONLIMIT) * (fov / basefov);
				bikeroll += DEVICE[player].XPOS / 10.0f * sensitivity * (fov / basefov) / 100;
				if(bikex < 0)
					bikex += BIKEXROTATIONLIMIT;
				else if(bikex >= BIKEXROTATIONLIMIT)
					bikex -= BIKEXROTATIONLIMIT;
				if(bikeroll < -BIKEROLLLIMIT)
					bikeroll = -BIKEROLLLIMIT;
				else if(bikeroll > BIKEROLLLIMIT)
					bikeroll = BIKEROLLLIMIT;
				EMU_WriteFloat(bikebase[player][0], 0, bikex);
				EMU_WriteFloat(bikebase[player][0], PD_bikeroll, bikeroll);
			}
			camy += (!PROFILE[player].SETTINGS[INVERTPITCH] ? -DEVICE[player].YPOS : DEVICE[player].YPOS) / 10.0f * sensitivity * (fov / basefov);
			if(camy < -90)
				camy = -90;
			else if(camy > 90)
				camy = 90;
			EMU_WriteFloat(playerbase[player], PD_camy, camy);
			if(PROFILE[player].SETTINGS[CROSSHAIR]) // if crosshair movement is enabled
			{
				float gunx = EMU_ReadFloat(playerbase[player], PD_gunrx), crosshairx = EMU_ReadFloat(playerbase[player], PD_crosshairx); // after camera x and y have been calculated and injected, calculate the gun/crosshair movement
				gunx += DEVICE[player].XPOS / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / basefov) * 0.05f;
				crosshairx += DEVICE[player].XPOS / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / 4 / (basefov / 4)) * 0.05f;
				if(gunx < -GUNAIMLIMIT)
					gunx = -GUNAIMLIMIT;
				else if(gunx > GUNAIMLIMIT)
					gunx = GUNAIMLIMIT;
				if(crosshairx < -CROSSHAIRLIMIT)
					crosshairx = -CROSSHAIRLIMIT;
				else if(crosshairx > CROSSHAIRLIMIT)
					crosshairx = CROSSHAIRLIMIT;
				EMU_WriteFloat(playerbase[player], PD_gunrx, gunx);
				EMU_WriteFloat(playerbase[player], PD_gunlx, gunx);
				EMU_WriteFloat(playerbase[player], PD_crosshairx, crosshairx);
				if(camy > -90 && camy < 90) // only allow player's gun to pitch within a valid range
				{
					float guny = EMU_ReadFloat(playerbase[player], PD_gunry), crosshairy = EMU_ReadFloat(playerbase[player], PD_crosshairy);
					guny += (!PROFILE[player].SETTINGS[INVERTPITCH] ? DEVICE[player].YPOS : -DEVICE[player].YPOS) / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / basefov) * 0.075f;
					crosshairy += (!PROFILE[player].SETTINGS[INVERTPITCH] ? DEVICE[player].YPOS : -DEVICE[player].YPOS) / (!aimingflag ? 10.0f : 40.0f) * gunsensitivity * (fov / 4 / (basefov / 4)) * 0.1f;
					if(guny < -GUNAIMLIMIT)
						guny = -GUNAIMLIMIT;
					else if(guny > GUNAIMLIMIT)
						guny = GUNAIMLIMIT;
					if(crosshairy < -CROSSHAIRLIMIT)
						crosshairy = -CROSSHAIRLIMIT;
					else if(crosshairy > CROSSHAIRLIMIT)
						crosshairy = CROSSHAIRLIMIT;
					EMU_WriteFloat(playerbase[player], PD_gunry, guny);
					EMU_WriteFloat(playerbase[player], PD_gunly, guny);
					EMU_WriteFloat(playerbase[player], PD_crosshairy, crosshairy);
				}
			}
		}
		else if(!dead && camera == 1 && !menu) // player is alive, in first person and is in menu
			PD_RadialMenuNav(player); // check if player is in radial menu and translate mouse input to C buttons
		else if(dead || camera != 1) // player is dead or not in first person camera
			PD_ResetCrouchToggle(player); // reset crouch toggle
	}
	PD_Controller(); // set controller data
}
//==========================================================================
// Purpose: crouching function for Perfect Dark (2 = stand, 0 = duck)
// Changes Globals: safetoduck, crouchstance, safetostand
//==========================================================================
static void PD_Crouch(const int player)
{
	int crouchheld = DEVICE[player].BUTTONPRIM[CROUCH] || DEVICE[player].BUTTONSEC[CROUCH];
	if(PROFILE[player].SETTINGS[CROUCHTOGGLE]) // check and change player stance
	{
		if(safetoduck[player] && crouchheld) // standing to ducking
			safetoduck[player] = 0, crouchstance[player] = 1;
		if(!safetoduck[player] && !crouchheld) // crouch is no longer being held, ready to stand
			safetostand[player] = 1;
		if(safetostand[player] && crouchheld) // standing up
			safetoduck[player] = 1, crouchstance[player] = 0;
		if(safetostand[player] && safetoduck[player] && !crouchheld) // crouch key not active, ready to change toggle
			safetostand[player] = 0;
		crouchheld = crouchstance[player];
	}
	if(crouchheld) // user is holding down the crouch button
		EMU_WriteInt(playerbase[player], 0, 0); // set in-game crouch flag to lowest point
	else // player isn't holding down crouch
		EMU_WriteInt(playerbase[player], 0, 2); // force standing flag value (the side effect to this brute force method is it causes a speed glitch while crawling in tunnels in PD only)
}
//==========================================================================
// Purpose: translate mouse to weapon radial menu
// Changes Globals: xmenu, ymenu, radialmenudirection
//==========================================================================
static void PD_RadialMenuNav(const int player)
{
	const float max = 19, threshold = 13, diagonalthres = 0.75;
	if(PROFILE[player].SETTINGS[PDRADIALNAV] && (DEVICE[player].BUTTONPRIM[ACCEPT] || DEVICE[player].BUTTONSEC[ACCEPT]) && !DEVICE[player].BUTTONPRIM[FIRE] && !DEVICE[player].BUTTONSEC[FIRE]) // if a button is held (reject if fire is pressed so aimx/y can be reset back to center)
	{
		xmenu[player] += DEVICE[player].XPOS / 10.0f * PROFILE[player].SETTINGS[SENSITIVITY] / 40.0f;
		ymenu[player] += DEVICE[player].YPOS / 10.0f * PROFILE[player].SETTINGS[SENSITIVITY] / 40.0f;
		if(xmenu[player] > max)
			xmenu[player] = max;
		else if(xmenu[player] < -max)
			xmenu[player] = -max;
		if(ymenu[player] > max)
			ymenu[player] = max;
		else if(ymenu[player] < -max)
			ymenu[player] = -max;
		if(ymenu[player] < -threshold) // c-up
			radialmenudirection[player][FORWARDS] = 1;
		else if(ymenu[player] > threshold) // c-down
			radialmenudirection[player][BACKWARDS] = 1;
		if(xmenu[player] < -threshold) // c-left
			radialmenudirection[player][STRAFELEFT] = 1;
		else if(xmenu[player] > threshold) // c-right
			radialmenudirection[player][STRAFERIGHT] = 1;
		if(xmenu[player] < -threshold && ymenu[player] < -threshold && (-xmenu[player] + -ymenu[player]) / (max * 2) > diagonalthres) // c-up-left
			radialmenudirection[player][FORWARDS] = 1, radialmenudirection[player][STRAFELEFT] = 1;
		else if(xmenu[player] > threshold && ymenu[player] < -threshold && (xmenu[player] + -ymenu[player]) / (max * 2) > diagonalthres) // c-up-right
			radialmenudirection[player][FORWARDS] = 1, radialmenudirection[player][STRAFERIGHT] = 1;
		else if(xmenu[player] < -threshold && ymenu[player] > threshold && (-xmenu[player] + ymenu[player]) / (max * 2) > diagonalthres) // c-down-left
			radialmenudirection[player][BACKWARDS] = 1, radialmenudirection[player][STRAFELEFT] = 1;
		else if(xmenu[player] > threshold && ymenu[player] > threshold && (xmenu[player] + ymenu[player]) / (max * 2) > diagonalthres) // c-down-right
			radialmenudirection[player][BACKWARDS] = 1, radialmenudirection[player][STRAFERIGHT] = 1;
	}
	else
		xmenu[player] = 0, ymenu[player] = 0;
}
//==========================================================================
// Purpose: calculate and send emulator key combo
// Changes Globals: radialmenudirection
//==========================================================================
static void PD_Controller(void)
{
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		Controller[player].U_CBUTTON = DEVICE[player].BUTTONPRIM[FORWARDS] || DEVICE[player].BUTTONSEC[FORWARDS] || radialmenudirection[player][FORWARDS];
		Controller[player].D_CBUTTON = DEVICE[player].BUTTONPRIM[BACKWARDS] || DEVICE[player].BUTTONSEC[BACKWARDS] || radialmenudirection[player][BACKWARDS];
		Controller[player].L_CBUTTON = DEVICE[player].BUTTONPRIM[STRAFELEFT] || DEVICE[player].BUTTONSEC[STRAFELEFT] || radialmenudirection[player][STRAFELEFT];
		Controller[player].R_CBUTTON = DEVICE[player].BUTTONPRIM[STRAFERIGHT] || DEVICE[player].BUTTONSEC[STRAFERIGHT] || radialmenudirection[player][STRAFERIGHT];
		Controller[player].Z_TRIG = DEVICE[player].BUTTONPRIM[FIRE] || DEVICE[player].BUTTONSEC[FIRE] || DEVICE[player].BUTTONPRIM[PREVIOUSWEAPON] || DEVICE[player].BUTTONSEC[PREVIOUSWEAPON];
		Controller[player].R_TRIG = DEVICE[player].BUTTONPRIM[AIM] || DEVICE[player].BUTTONSEC[AIM];
		Controller[player].A_BUTTON = DEVICE[player].BUTTONPRIM[ACCEPT] || DEVICE[player].BUTTONSEC[ACCEPT] || DEVICE[player].BUTTONPRIM[PREVIOUSWEAPON] || DEVICE[player].BUTTONSEC[PREVIOUSWEAPON] || DEVICE[player].BUTTONPRIM[NEXTWEAPON] || DEVICE[player].BUTTONSEC[NEXTWEAPON];
		Controller[player].B_BUTTON = DEVICE[player].BUTTONPRIM[CANCEL] || DEVICE[player].BUTTONSEC[CANCEL];
		Controller[player].START_BUTTON = DEVICE[player].BUTTONPRIM[START] || DEVICE[player].BUTTONSEC[START];
		DEVICE[player].ARROW[0] = (DEVICE[player].BUTTONPRIM[UP] || DEVICE[player].BUTTONSEC[UP]) ? 120 : 0;
		DEVICE[player].ARROW[1] = (DEVICE[player].BUTTONPRIM[DOWN] || DEVICE[player].BUTTONSEC[DOWN]) ? -120 : 0;
		DEVICE[player].ARROW[2] = (DEVICE[player].BUTTONPRIM[LEFT] || DEVICE[player].BUTTONSEC[LEFT]) ? -120 : 0;
		DEVICE[player].ARROW[3] = (DEVICE[player].BUTTONPRIM[RIGHT] || DEVICE[player].BUTTONSEC[RIGHT]) ? 120 : 0;
		Controller[player].X_AXIS = DEVICE[player].ARROW[0] + DEVICE[player].ARROW[1];
		Controller[player].Y_AXIS = DEVICE[player].ARROW[2] + DEVICE[player].ARROW[3];
		PD_ResetRadialMenuBtns(player); // reset radialmenudirection
	}
}
//==========================================================================
// Purpose: change default fov
//==========================================================================
static void PD_FOVHack(void)
{
	float newfov = overridefov;
	unsigned int unsignedinteger = *(unsigned int *)(float *)(&newfov);
	EMU_WriteInt(PD_defaultfov, 0, 0x3C010000 + (short)(unsignedinteger / 0x10000));
}
//==========================================================================
// Purpose: force PD to unlock fps
// Changes Globals: limitspeedhack
// Q: Why only inject the speed-hack if PD is running at 30fps?
// A: If PD is running lower than 30fps it must be because there is something taxing the graphics and would cause extra stuttering if you try and combat the natural slowdown. Not even an overclock emulator can overcome the few situations with too many characters on screen - so just wait it out until the game goes back to 30fps
//==========================================================================
static void PD_SpeedHack(void)
{
	if(limitspeedhack < 200)
		limitspeedhack++;
	else
		limitspeedhack = 0;
	if(limitspeedhack % 10 == 0) // do not inject too quickly (causes instability)
	{
		unsigned int slowmotioncheat = (unsigned int)EMU_ReadInt(PD_slowmocheat, 0);
		if(slowmotioncheat < 0xC0000000 && slowmotioncheat > 0x7F000000 || slowmotioncheat < 0x40000000) // if slo-mo single player cheat is off
			slowmotioncheat = 0;
		else
			slowmotioncheat = 1; // slow motion cheat is on
		if((unsigned int)EMU_ReadInt(PD_officialrom, 0) != 0x4E504445) // if playing romhack disable slo-mo single player cheat check (GoldenEye: X replaced cheat with Turbo Mode)
			slowmotioncheat = 0;
		if(EMU_ReadInt(PD_tickrate, 0) == 2 && EMU_ReadInt(PD_speed, 0) == 0 && (unsigned int)EMU_ReadInt(PD_mpspeed, 0) < 0x40000000 && slowmotioncheat == 0) // if PD is running at 30fps, the player isn't using combat boost, mp slow motion setting is off and sp slow motion cheat is off
		{
			for(int bruteforce = 0; bruteforce < 48; bruteforce++)
			{
				EMU_WriteInt(PD_timer, 0, 0x00000000); // trick the system into thinking there is plenty of room left to run at full speed
				if(bruteforce % 40 == 0)
					DEV_Sleep(1);
			}
			EMU_WriteInt(PD_timer, 0, 0x00010000); // set clock back to default (reset fps lock to 60)
		}
	}
}
//==========================================================================
// Purpose: run when emulator closes rom
// Changes Globals: playerbase, bikebase, safetoduck, safetostand, crouchstance, xmenu, ymenu, limitspeedhack
//==========================================================================
static void PD_Quit(void)
{
	for(int player = PLAYER1; player < ALLPLAYERS; player++)
	{
		playerbase[player] = 0;
		bikebase[player][0] = 0;
		bikebase[player][1] = 0;
		PD_ResetCrouchToggle(player);
		xmenu[player] = 0;
		ymenu[player] = 0;
	}
	limitspeedhack = 0;
}