//==========================================================================
// Mouse Injector Plugin
//==========================================================================
// Copyright (C) 2017 Carnivorous
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the <organization> nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL CARNIVOROUS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//==========================================================================
#include "memory.h"

//==========================================================================
// Purpose: read short from memory and then return memory value
// Parameter: address location plus offset
//==========================================================================
short EMU_ReadShort(const unsigned int addr, const unsigned int offset)
{
	return rdramptr != 0 && addr / 0x1000000 == 0x80 ? LOAD_SHALF_PARAM(addr + offset + 2) : (short)0xABAD; // return garbage if request invalid location
}
//==========================================================================
// Purpose: read int from memory and then return memory value
// Parameter: address location plus offset
//==========================================================================
int EMU_ReadInt(const unsigned int addr, const unsigned int offset)
{
	return rdramptr != 0 && addr / 0x1000000 == 0x80 ? LOAD_SWORD_PARAM(addr + offset) : (int)0xABADC0DE;
}
//==========================================================================
// Purpose: read float from memory and then return memory value
// Parameter: address location plus offset
//==========================================================================
float EMU_ReadFloat(const unsigned int addr, const unsigned int offset)
{
	return rdramptr != 0 && addr / 0x1000000 == 0x80 ? *((float *)&LOAD_UWORD_PARAM(addr + offset)) : (int)0xABADC0DE;
}
//==========================================================================
// Purpose: read int from emulator's copy of rom and then return memory value
// Parameter: address location plus offset
//==========================================================================
int EMU_ReadROM(const unsigned int addr)
{
	return (int)romptr[addr / 0x4];
}
//==========================================================================
// Purpose: write short to memory location
// Parameter: address location, offset and value
//==========================================================================
void EMU_WriteShort(const unsigned int addr, const unsigned int offset, const short value)
{
	if(rdramptr != 0 && addr / 0x1000000 == 0x80)
		LOAD_SHALF_PARAM(addr + offset + 2) = value;
}
//==========================================================================
// Purpose: write int to memory location
// Parameter: address location, offset and value
//==========================================================================
void EMU_WriteInt(const unsigned int addr, const unsigned int offset, const int value)
{
	if(rdramptr != 0 && addr / 0x1000000 == 0x80)
		LOAD_SWORD_PARAM(addr + offset) = value;
}
//==========================================================================
// Purpose: write float to memory location
// Parameter: address location, offset and value
//==========================================================================
void EMU_WriteFloat(const unsigned int addr, const unsigned int offset, const float value)
{
	if(rdramptr != 0 && addr / 0x1000000 == 0x80)
		LOAD_UWORD_PARAM(addr + offset) = *(unsigned int *)(&value);
}
//==========================================================================
// Purpose: write to emulator's copy of rom
// Parameter: address location and value
//==========================================================================
void EMU_WriteROM(const unsigned int addr, const unsigned int value)
{
	if(romptr != 0)
		romptr[addr / 0x4] = (void *)value;
}